from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, List, Type, Union

if TYPE_CHECKING:
    from ..interfaces.dispatcher import DispatcherInterface


class BaseDispatcher(metaclass=ABCMeta):
    """所有非单函数型 Dispatcher 的基类, 用于为参数解析提供可扩展的支持."""

    mixin: List[Union["BaseDispatcher", Type["BaseDispatcher"]]]
    """声明该 Dispatcher 所包含的来自其他 Dispatcher 提供的参数解析支持,
    若某参数该 Dispatcher 无法解析, 将跳转到该列表中并交由其中的 Dispatcher 进行解析,
    该列表中的 Dispatcher 全部被调用过且都不返回一有效值时才会将解析权交由其他的 Dispatcher.
    """

    @abstractmethod
    async def catch(self, interface: "DispatcherInterface"):
        """该方法可以是 `staticmethod`, `classmethod` 亦或是普通的方法/函数.
        唯一的要求是 `Dispatcher.catch` 获取到的必须为一可调用异步 Callable.

        Args:
            interface (DispatcherInterface): `Dispatcher` 服务的主要对象, 可以从其中获取以下信息:
             - 当前解析中的参数的信息;
             - 当前执行的信息, 比如正在处理的事件, `Listener`/`ExecTarget` etc.;
        """
        pass

    beforeExecution = None
    afterDispatch = None
    afterExecution = None



from typing import TYPE_CHECKING, Any, Dict

from ..entities.decorator import Decorator
from ..entities.dispatcher import BaseDispatcher
from ..entities.signatures import Force
from ..utilles import Ctx, run_always_await_safely

if TYPE_CHECKING:
    from ..interfaces.dispatcher import DispatcherInterface


ctx_dei_returnvalue = Ctx("ctx_dei_returnvalue")


class DecoratorInterface(BaseDispatcher):
    """Broadcast Control 内部机制 Decorator 的具体管理实现"""

    dispatcher_interface: "DispatcherInterface"

    def __init__(self, dispatcher_interface: "DispatcherInterface"):
        self.dispatcher_interface = dispatcher_interface

    @property
    def name(self):
        return self.dispatcher_interface.name

    @property
    def annotation(self):
        return self.dispatcher_interface.annotation

    @property
    def event(self):
        return self.dispatcher_interface.event

    @property
    def return_value(self):
        return ctx_dei_returnvalue.get()

    @property
    def local_storage(self):
        return self.dispatcher_interface.execution_contexts[-1].local_storage

    async def catch(self, interface: "DispatcherInterface"):
        if isinstance(interface.default, Decorator):
            decorator: Decorator = interface.default
            with ctx_dei_returnvalue.use(
                await interface.lookup_param(interface.name, interface.annotation, None, [[], 0])
                if not decorator.pre
                else None
            ):
                return Force(await run_always_await_safely(decorator.target, self))



from typing import Any, Callable, Dict, List, TypeVar

from graia.broadcast.utilles import NestableIterable

from ..typing import T_Dispatcher

T = TypeVar("T")
I = TypeVar("I")


class ExecutionContext:
    __slots__ = ("event", "dispatchers", "path", "local_storage")

    dispatchers: List[T_Dispatcher]
    path: NestableIterable
    local_storage: Dict[str, Any]

    def __init__(self, dispatchers: List[T_Dispatcher]):
        self.dispatchers = dispatchers
        self.local_storage = {}



from typing import TYPE_CHECKING, Any, Callable, Type, Union

if TYPE_CHECKING:
    from graia.broadcast.entities.dispatcher import BaseDispatcher
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface


T_Dispatcher = Union[Type["BaseDispatcher"], "BaseDispatcher", Callable[["DispatcherInterface"], Any]]

T_Dispatcher_Callable = Callable[["DispatcherInterface"], Any]



from abc import ABCMeta, abstractmethod
from types import TracebackType
from typing import List, Optional, Type, Union

from ..interfaces.dispatcher import DispatcherInterface as DispatcherInterface

class BaseDispatcher(metaclass=ABCMeta):
    mixin: List[Union[BaseDispatcher, Type[BaseDispatcher]]]
    @abstractmethod
    async def catch(self, interface: DispatcherInterface): ...
    def beforeExecution(self, interface: DispatcherInterface): ...
    def afterDispatch(
        self,
        interface: DispatcherInterface,
        exception: Optional[Exception],
        tb: Optional[TracebackType],
    ): ...
    def afterExecution(
        self,
        interface: DispatcherInterface,
        exception: Optional[Exception],
        tb: Optional[TracebackType],
    ): ...



from typing import TYPE_CHECKING, Any, Generic, List, Tuple, TypeVar

from ..entities.context import ExecutionContext
from ..entities.dispatcher import BaseDispatcher
from ..entities.event import Dispatchable
from ..entities.signatures import Force
from ..exceptions import RequirementCrashed
from ..typing import T_Dispatcher
from ..utilles import NestableIterable

if TYPE_CHECKING:
    from .. import Broadcast


class EmptyEvent(Dispatchable):
    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(_):
            pass


T_Event = TypeVar("T_Event", bound=Dispatchable)


class DispatcherInterface(Generic[T_Event]):
    broadcast: "Broadcast"

    execution_contexts: List[ExecutionContext]
    parameter_contexts: List[Tuple[str, Any, Any]]

    def __init__(self, broadcast_instance: "Broadcast") -> None:
        self.broadcast = broadcast_instance
        self.execution_contexts = []
        self.parameter_contexts = []

    @property
    def name(self) -> str:
        return self.parameter_contexts[-1][0]

    @property
    def annotation(self) -> Any:
        return self.parameter_contexts[-1][1]

    @property
    def default(self) -> Any:
        return self.parameter_contexts[-1][2]

    @property
    def event(self) -> T_Event:
        return self.broadcast.event_ctx.get()  # type: ignore

    @property
    def global_dispatcher(self) -> List[T_Dispatcher]:
        return self.broadcast.global_dispatchers

    def start_execution(
        self,
        dispatchers: List[T_Dispatcher],
    ):
        self.execution_contexts.append(ExecutionContext(dispatchers))
        self.execution_contexts[-1].path = NestableIterable([])
        return self

    def clean(self):
        self.execution_contexts.pop()

    def inject_execution_raw(self, *dispatchers: T_Dispatcher):
        for dispatcher in dispatchers:
            self.execution_contexts[-1].dispatchers.insert(0, dispatcher)

    def inject_global_raw(self, *dispatchers: T_Dispatcher):
        for dispatcher in dispatchers[::-1]:
            self.broadcast.global_dispatchers.insert(1, dispatcher)

    async def lookup_param(
        self,
        name: str,
        annotation: Any,
        default: Any,
        optimized_log: Any,
    ) -> Any:
        self.parameter_contexts.append((name, annotation, default))
        dispatch_path = self.execution_contexts[-1].path

        try:
            dispatch_path.iterable = optimized_log
            for dispatcher in dispatch_path:
                result = await getattr(dispatcher, "catch", dispatcher)(self)  # type: ignore
                if result is None:  # 不可靠.
                    break
                if result.__class__ is Force:
                    return result.target
                return result
            optimized_log.clear()
            dispatch_path.iterable = self.execution_contexts[-1].dispatchers
            for dispatcher in dispatch_path:
                result = await getattr(dispatcher, "catch", dispatcher)(self)  # type: ignore

                if result is None:
                    continue

                if result.__class__ is Force:
                    return result.target

                optimized_log.append(dispatcher)
                return result
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            self.parameter_contexts.pop()

    async def lookup_by_directly(self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any) -> Any:
        self.parameter_contexts.append((name, annotation, default))

        try:
            result = await getattr(dispatcher, "catch", dispatcher)(self)  # type: ignore
            if result.__class__ is Force:
                return result.target

            return result
        finally:
            self.parameter_contexts.pop()



import asyncio
import sys
import traceback
from collections import UserList
from typing import Callable, Dict, Iterable, List, Optional, Type, Union

from .builtin.event import ExceptionThrowed
from .entities.decorator import Decorator
from .entities.event import Dispatchable
from .entities.exectarget import ExecTarget
from .entities.listener import Listener
from .entities.namespace import Namespace
from .entities.signatures import Force, RemoveMe
from .exceptions import (DisabledNamespace, ExecutionStop, ExistedNamespace,
                         InvalidEventName, PropagationCancelled,
                         RegisteredEventListener, RequirementCrashed,
                         UnexistedNamespace)
from .interfaces.decorator import DecoratorInterface
from .interfaces.dispatcher import DispatcherInterface
from .typing import T_Dispatcher
from .utilles import (Ctx, argument_signature, dispatcher_mixin_handler,
                      group_dict, run_always_await_safely)


class DebugList(UserList):
    def extend(self, item) -> None:
        print(item)
        return super().extend(item)

    def append(self, item) -> None:
        print(item)
        return super().append(item)

    def insert(self, i: int, item) -> None:
        print(i, item)
        return super().insert(i, item)


class Broadcast:
    loop: asyncio.AbstractEventLoop

    default_namespace: Namespace
    namespaces: List[Namespace]
    listeners: List[Listener]

    dispatcher_interface: DispatcherInterface
    decorator_interface: DecoratorInterface

    event_ctx: Ctx[Dispatchable]

    debug_flag: bool

    global_dispatchers: List["T_Dispatcher"]

    def __init__(
        self,
        *,
        loop: asyncio.AbstractEventLoop = None,
        debug_flag: bool = False,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.default_namespace = Namespace(name="default", default=True)
        self.debug_flag = debug_flag
        self.namespaces = []
        self.listeners = []
        self.event_ctx = Ctx("bcc_event_ctx")
        self.dispatcher_interface = DispatcherInterface(self)
        self.decorator_interface = DecoratorInterface(self.dispatcher_interface)
        self.global_dispatchers = [self.decorator_interface]

        @self.dispatcher_interface.inject_global_raw
        async def _(interface: DispatcherInterface):
            if interface.annotation is interface.event.__class__:
                return interface.event
            elif interface.annotation is Broadcast:
                return interface.broadcast
            elif interface.annotation is DispatcherInterface:
                return interface

    def default_listener_generator(self, event_class) -> Iterable[Listener]:
        return list(
            filter(
                lambda x: all(
                    [
                        not x.namespace.hide,
                        not x.namespace.disabled,
                        event_class in x.listening_events,
                    ]
                ),
                self.listeners,
            )
        )

    async def layered_scheduler(self, listener_generator: Iterable[Listener], event: Dispatchable):
        grouped: Dict[int, List[Listener]] = group_dict(listener_generator, lambda x: x.priority)
        event_dispatcher_mixin = dispatcher_mixin_handler(event.Dispatcher)
        with self.event_ctx.use(event):
            for _, current_group in sorted(grouped.items(), key=lambda x: x[0]):
                coros = [self.Executor(target=i, dispatchers=event_dispatcher_mixin) for i in current_group]
                done_tasks, _ = await asyncio.wait(coros)
                for task in done_tasks:
                    if task.exception().__class__ is PropagationCancelled:
                        break

    async def Executor(
        self,
        target: Union[Callable, ExecTarget],
        dispatchers: List[T_Dispatcher] = None,
        post_exception_event: bool = True,
        print_exception: bool = True,
        use_global_dispatchers: bool = True,
    ):
        is_exectarget = is_listener = False
        if isinstance(target, Listener):
            is_exectarget = is_listener = True
        elif isinstance(target, ExecTarget):
            is_exectarget = True

        if is_listener:
            if target.namespace.disabled:  # type: ignore
                raise DisabledNamespace(
                    "caught a disabled namespace: {0}".format(target.namespace.name)  # type: ignore
                )

        target_callable = target.callable if is_exectarget else target  # type: ignore
        parameter_compile_result = {}

        dii = self.dispatcher_interface

        dispatchers = dispatchers and dispatchers.copy() or []

        if is_exectarget:
            dispatchers.extend(target.dispatchers)
            if is_listener:
                dispatchers.extend(target.namespace.injected_dispatchers)  # type: ignore

        if use_global_dispatchers:
            for i in self.global_dispatchers:
                dispatchers.insert(0, i)

        dii.start_execution(dispatchers)
        try:
            for dispatcher in dii.execution_contexts[-1].dispatchers:
                i = getattr(dispatcher, "beforeExecution", None)
                if i:
                    await run_always_await_safely(i, dii)  # type: ignore

            if is_exectarget:
                for name, annotation, default in argument_signature(target_callable):  # type: ignore
                    optimized_log = target.param_paths.setdefault(name, [])
                    parameter_compile_result[name] = await dii.lookup_param(name, annotation, default, optimized_log)

                for hl_d in target.decorators:
                    await dii.lookup_by_directly(
                        self.decorator_interface,
                        "_bcc_headless_decorators",
                        None,
                        hl_d,
                    )

            else:
                for name, annotation, default in argument_signature(target_callable):  # type: ignore
                    parameter_compile_result[name] = await dii.lookup_param(name, annotation, default, [])

            for dispatcher in dii.execution_contexts[-1].dispatchers:
                i = getattr(dispatcher, "afterDispatch", None)
                if i:
                    await run_always_await_safely(i, dii, None, None)  # type: ignore

            result = await run_always_await_safely(target_callable, **parameter_compile_result)
        except (ExecutionStop, PropagationCancelled):
            raise
        except RequirementCrashed:
            traceback.print_exc()
            raise
        except Exception as e:
            event: Optional[Dispatchable] = self.event_ctx.get()
            if event is not None and event.__class__ is not ExceptionThrowed:
                if print_exception:
                    traceback.print_exc()
                if post_exception_event:
                    self.postEvent(ExceptionThrowed(exception=e, event=event))
            raise
        finally:
            _, exception, tb = sys.exc_info()
            for dispatcher in dii.execution_contexts[-1].dispatchers:
                i = getattr(dispatcher, "afterExecution", None)
                if i:
                    await run_always_await_safely(i, dii, exception, tb)  # type: ignore

            dii.clean()

        if result.__class__ is Force:
            return result.content
        elif result is RemoveMe:
            if is_listener and target in self.listeners:
                self.listeners.remove(target)
        return result

    def postEvent(self, event: Dispatchable):
        self.loop.create_task(
            self.layered_scheduler(
                listener_generator=self.default_listener_generator(event.__class__),
                event=event,
            )
        )

    @staticmethod
    def event_class_generator(target=Dispatchable):
        for i in target.__subclasses__():
            yield i
            if i.__subclasses__():
                yield from Broadcast.event_class_generator(i)

    @staticmethod
    def findEvent(name: str):
        for i in Broadcast.event_class_generator():
            if i.__name__ == name:
                return i

    def getDefaultNamespace(self):
        return self.default_namespace

    def createNamespace(self, name, *, priority: int = 0, hide: bool = False, disabled: bool = False):
        if self.containNamespace(name):
            raise ExistedNamespace(name, "has been created!")
        self.namespaces.append(Namespace(name=name, priority=priority, hide=hide, disabled=disabled))
        return self.namespaces[-1]

    def removeNamespace(self, name):
        if self.containNamespace(name):
            for index, i in enumerate(self.namespaces):
                if i.name == name:
                    self.namespaces.pop(index)
                    return
        else:
            raise UnexistedNamespace(name)

    def containNamespace(self, name):
        for i in self.namespaces:
            if i.name == name:
                return True
        return False

    def getNamespace(self, name) -> "Namespace":
        if self.containNamespace(name):
            for i in self.namespaces:
                if i.name == name:
                    return i
            else:
                raise UnexistedNamespace(name)
        else:
            raise UnexistedNamespace(name)

    def hideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = True

    def unhideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = False

    def disableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = True

    def enableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = False

    def containListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return True
        return False

    def getListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return i

    def removeListener(self, target):
        self.listeners.remove(target)

    def receiver(
        self,
        event: Union[str, Type[Dispatchable]],
        priority: int = 16,
        dispatchers: List[T_Dispatcher] = None,
        namespace: Namespace = None,
        decorators: List[Decorator] = None,
    ):
        if isinstance(event, str):
            _name = event
            event = self.findEvent(event)  # type: ignore
            if not event:
                raise InvalidEventName(_name + " is not valid!")  # type: ignore
        priority = int(priority)

        def receiver_wrapper(callable_target):
            may_listener = self.getListener(callable_target)
            if not may_listener:
                self.listeners.append(
                    Listener(
                        callable=callable_target,
                        namespace=namespace or self.getDefaultNamespace(),
                        inline_dispatchers=dispatchers,
                        priority=priority,
                        listening_events=[event],  # type: ignore
                        decorators=decorators,
                    )
                )
            else:
                if event not in may_listener.listening_events:
                    may_listener.listening_events.append(event)  # type: ignore
                else:
                    raise RegisteredEventListener(event.__name__, "has been registered!")  # type: ignore
            return callable_target

        return receiver_wrapper



from typing import Callable, Dict, List

from ..typing import T_Dispatcher
from .decorator import Decorator


class ExecTarget:
    callable: Callable
    dispatchers: List[T_Dispatcher]
    decorators: List[Decorator]

    param_paths: Dict[str, List[T_Dispatcher]]

    def __init__(
        self,
        callable: Callable,
        inline_dispatchers: List[T_Dispatcher] = None,
        decorators: List[Decorator] = None,
    ):
        self.callable = callable
        self.dispatchers = inline_dispatchers or []
        self.decorators = decorators or []

        self.param_paths = {}
